"""Test querying with properties."""
import pytest
from .fixtures import fixture_neo4j_driver
from reasoner_transpiler.attributes import set_custom_attribute_types, set_custom_attribute_skip_list, \
    reset_custom_attribute_types, DEFAULT_ATTRIBUTE_TYPE
from reasoner_transpiler.cypher import get_query


def test_numeric(neo4j_driver):
    """Test querying with numeric property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Gene",
                "constraints": [
                    {"id": "length",
                     "value": 277,
                     "operator": "==="}
                ]
            },
        },
        "edges": {},
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    results = sorted(
        output["knowledge_graph"]["nodes"].values(),
        key=lambda node: node["name"],
    )
    expected_nodes = [
        "CASP3",
    ]
    for ind, result in enumerate(results):
        assert result["name"] == expected_nodes[ind]


def test_string(neo4j_driver):
    """Test querying with string property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Gene",
                "constraints": [
                    {"id": "chromosome",
                     "value": "17",
                     "operator": "==="}
                ]
            },
        },
        "edges": {},
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    results = sorted(
        output["knowledge_graph"]["nodes"].values(),
        key=lambda node: node["name"],
    )
    expected_nodes = [
        "BRCA1",
    ]
    for ind, result in enumerate(results):
        assert result["name"] == expected_nodes[ind]


def test_bool(neo4j_driver):
    """Test querying with boolean property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:ChemicalSubstance",
            },
            "n1": {
                "categories": "biolink:Disease",
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:treats",
                "attribute_constraints": [
                    {"id": "fda_approved",
                     "value": True,
                     "operator": "==="}
                ]
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 1
    results = sorted(
        output["knowledge_graph"]["nodes"].values(),
        key=lambda node: node["name"],
    )
    expected_nodes = [
        "metformin", "type 2 diabetes mellitus",
    ]
    for ind, result in enumerate(results):
        assert result["name"] == expected_nodes[ind]


def test_publications(neo4j_driver):
    """Test publications."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:836"],
            },
            "n1": {
                "ids": ["NCBIGene:841"],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    attributes = list(edges.values())[0]["attributes"]
    assert len(attributes) == 1
    assert attributes[0] == {
        "original_attribute_name": "publications",
        "attribute_type_id": "biolink:publications",
        "value": ["xxx"],
        "value_type_id": "linkml:Uriorcurie"
    }


def test_empty_constraints(neo4j_driver):
    """Test querying with empty 'constraints' property."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Gene",
                "constraints": [],
            },
            "n1": {
                "constraints": [],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "attribute_constraints": [],
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 10


def test_valid_biolink_attribute_without_mapping(neo4j_driver):
    """publications is included in the default attribute types mapping included in the codebase,
    but here we test that it still gets recognized as being biolink compliant when it's not in the mapping"""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["MONDO:0005148"],
            },
            "n1": {
                "ids": ["NCBIGene:841"],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    attributes = list(edges.values())[0]["attributes"]
    assert len(attributes) == 1
    assert attributes[0] == {
        "original_attribute_name": "p_value",
        "attribute_type_id": "biolink:p_value",
        "value": 0.000007,
        "value_type_id": DEFAULT_ATTRIBUTE_TYPE['value_type_id']
    }
    reset_custom_attribute_types()


def test_invalid_biolink_attribute_without_mapping(neo4j_driver):
    """Test to make sure an attribute that isn't biolink compliant, and isn't custom mapped, returns correctly."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:836"]
            },
            "n1": {
                "ids": ["MONDO:0005148"]
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1"
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    attribute = list(edges.values())[0]["attributes"][0]
    expected_attribute = {
        "original_attribute_name": "non_biolink_attribute",
        "attribute_type_id": DEFAULT_ATTRIBUTE_TYPE['attribute_type_id'],
        "value": "xxx123",
        "value_type_id": DEFAULT_ATTRIBUTE_TYPE['value_type_id']
    }
    assert attribute == expected_attribute


def test_json_attributes(neo4j_driver):
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:672"]
            },
            "n1": {
                "ids": ["MONDO:0004993"]
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
                "predicates": "biolink:gene_associated_with_condition"
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    correct_attributes = 0
    attributes = list(edges.values())[0]["attributes"]
    for attribute in attributes:
        if attribute.get("original_attribute_name", None) == "publications":
            assert attribute["value"] == ["xxx"]
            correct_attributes += 1
        elif attribute["attribute_type_id"] == "json_attribute_1":
            assert attribute["value"] == "json_value_1"
            correct_attributes += 1
        elif attribute["attribute_type_id"] == "json_attribute_2":
            correct_attributes += 1
        elif attribute["attribute_type_id"] == "json_attribute_3":
            assert attribute["value"] == "json_value_3"
            nested_attribute = attribute["attributes"][0]
            assert nested_attribute["attribute_type_id"] == "nested_json_attribute_1"
            assert nested_attribute["value"] == "nested_json_value_1"
            correct_attributes += 1
    assert correct_attributes == 4


def test_props_customization(neo4j_driver):
    qgraph = {
        "nodes": {
            "n0": {
                "ids": ["NCBIGene:836"],
            },
            "n1": {
                "ids": ["NCBIGene:841"],
            },
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }

    # put publications in the skip list and make sure it doesn't get included
    set_custom_attribute_skip_list(['publications'])
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    attributes = list(edges.values())[0]["attributes"]
    assert len(attributes) == 0

    # reset the skip list
    set_custom_attribute_skip_list([])

    # set custom attribute and value types for publications and make sure they get used instead of biolink or defaults
    set_custom_attribute_types({
        "publications": {
            "attribute_type_id": "transpiler:custom_attribute_type",
            "value_type_id": "transpiler:custom_value_type"
        }
    })
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    edges = output["knowledge_graph"]["edges"]
    assert len(edges) == 1
    attributes = list(edges.values())[0]["attributes"]
    assert len(attributes) == 1
    assert attributes[0] == {
        "original_attribute_name": "publications",
        "attribute_type_id": "transpiler:custom_attribute_type",
        "value": ["xxx"],
        "value_type_id": "transpiler:custom_value_type"
    }
    reset_custom_attribute_types()



import os
from bmt import Toolkit


BIOLINK_MODEL_VERSION = os.environ.get('BL_VERSION', '4.2.1')
BIOLINK_MODEL_SCHEMA_URL = f"https://raw.githubusercontent.com/biolink/biolink-model/v{BIOLINK_MODEL_VERSION}/biolink-model.yaml"
PREDICATE_MAP_URL = f"https://raw.githubusercontent.com/biolink/biolink-model/v{BIOLINK_MODEL_VERSION}/predicate_mapping.yaml"

bmt = Toolkit(schema=BIOLINK_MODEL_SCHEMA_URL, predicate_map=PREDICATE_MAP_URL)
ALL_BIOLINK_ENUMS = bmt.view.all_enums().keys()



import json
from pathlib import Path

from .biolink import bmt

DIR_PATH = Path(__file__).parent


def get_attribute_types_from_config():
    with open(DIR_PATH / "attribute_types.json", "r") as stream:
        return json.load(stream)


ATTRIBUTE_TYPES = get_attribute_types_from_config()

DEFAULT_ATTRIBUTE_TYPE = {
    "attribute_type_id": "biolink:Attribute",
    "value_type_id": "EDAM:data_0006",
}

ATTRIBUTE_SKIP_LIST = []

RESERVED_NODE_PROPS = [
    "id",
    "name",
    "labels",
    "element_id"
]
RESERVED_EDGE_PROPS = [
    "id",
    "predicate",
    "object",
    "subject",
    "sources"
]

# this should really be one representation or the other, or be configurable,
# but we have graphs with each now so temporarily (I hope, hope, hope) looking for both
EDGE_SOURCE_PROPS = [
    "aggregator_knowledge_source",
    "primary_knowledge_source",
    "biolink:aggregator_knowledge_source",
    "biolink:primary_knowledge_source"
]


def transform_attributes(result_item, node=False):

    # make a list of attributes to ignore while processing
    ignore_list = RESERVED_NODE_PROPS if node else EDGE_SOURCE_PROPS + RESERVED_EDGE_PROPS
    ignore_list += ATTRIBUTE_SKIP_LIST

    # an "attributes" attribute in neo4j should be a list of json strings,
    # attempt to start the attributes section of transformed attributes with its contents,
    # here we are assuming the attributes in "attributes" are already valid trapi
    json_attributes = []
    json_attributes_attribute = result_item.pop('attributes', None)
    if json_attributes_attribute:
        if isinstance(json_attributes_attribute, list):
            try:
                json_attributes = [json.loads(json_attribute_string)
                                   for json_attribute_string in json_attributes_attribute]
            except json.JSONDecodeError:
                print(f'!!! JSONDecodeError while parsing attributes property, ignoring: {json_attributes_attribute}')
        else:
            print(f'!!! the attributes edge property should be a list, ignoring: {json_attributes_attribute}')
    transformed_attributes = {
        'attributes': json_attributes
    }

    if not node:
        # for edges, find and format attributes that are qualifiers
        qualifiers = [key for key in result_item if key not in ignore_list
                      and bmt.is_qualifier(key)]
        transformed_attributes['qualifiers'] = [
            {"qualifier_type_id": f"biolink:{key}",
             "qualifier_value": value}
            for key, value in result_item.items() if key in qualifiers
        ]
    else:
        qualifiers = []

    # for attributes that aren't in ATTRIBUTE_TYPES, see if they are valid biolink attributes
    # add them to ATTRIBUTE_TYPES, so we don't need to look again
    for attribute in \
            [key for key in result_item.keys() if key not in ignore_list + qualifiers + list(ATTRIBUTE_TYPES.keys())]:
        attribute_mapping = DEFAULT_ATTRIBUTE_TYPE
        bmt_element = bmt.get_element(attribute)
        if bmt_element:
            if 'slot_uri' in bmt_element:
                attribute_mapping['attribute_type_id'] = bmt_element['slot_uri']
            elif 'class_uri' in bmt_element:
                attribute_mapping['attribute_type_id'] = bmt_element['class_uri']
        ATTRIBUTE_TYPES[attribute] = attribute_mapping

    # format the rest of the attributes, look up their attribute type and value type
    transformed_attributes['attributes'].extend([
        {'original_attribute_name': key,
         'value': value,
         # the following function will return
         # 'attribute_type_id': 'biolink-ified attribute type id'
         # 'value_type_id': 'biolink-ified value type id'
         **ATTRIBUTE_TYPES.get(key)}
        for key, value in result_item.items()
        if key not in ignore_list + qualifiers
    ])
    return transformed_attributes


def set_custom_attribute_types(attribute_types: dict):
    global ATTRIBUTE_TYPES
    ATTRIBUTE_TYPES = attribute_types


def set_custom_attribute_skip_list(skip_list: list):
    global ATTRIBUTE_SKIP_LIST
    ATTRIBUTE_SKIP_LIST = skip_list


def reset_custom_attribute_types():
    global ATTRIBUTE_TYPES
    ATTRIBUTE_TYPES = get_attribute_types_from_config()




"""Test query graph formats."""
import pytest
from reasoner_transpiler.cypher import get_query
from .fixtures import fixture_neo4j_driver


def test_curie_formats(neo4j_driver):
    """Test unusual curie formats."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": [
                    "MONDO:0005148",
                    "MONDO:0011122",
                ],
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:ChemicalSubstance",
            },
        },
        "edges": {
            "e01": {
                "predicates": [
                    "biolink:treats",
                ],
                "subject": "n1",
                "object": "n0",
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 5
    results = sorted(
        output["knowledge_graph"]["nodes"].values(),
        key=lambda node: node["name"],
    )
    expected_nodes = [
        "anagliptin",
        "bezafibrate",
        "metformin",
        "obesity disorder",
        "type 2 diabetes mellitus",
    ]
    for ind, result in enumerate(results):
        assert result["name"] == expected_nodes[ind]


def test_predicate_list():
    """Test that an edge with a list of predicates works properly."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:PhenotypicFeature",
            },
        },
        "edges": {
            "e01": {
                "predicates": ["biolink:capable_of", "biolink:biomarker_for"],
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    clause = get_query(qgraph, reasoner=False)
    # edges with types should be directed
    assert "(`n0`:`biolink:Disease`)-[`e01`:`biolink:biomarker_for`|`biolink:capable_of`]->(`n1`:`biolink:PhenotypicFeature`)" in clause


def test_single_edge_type_list():
    """Test that an edge with a edge type list of one works properly."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:PhenotypicFeature",
            },
        },
        "edges": {
            "e01": {
                "predicates": ["biolink:capable_of"],
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    clause = get_query(qgraph, reasoner=False)
    # edges with types should be directed
    assert "(`n0`:`biolink:Disease`)-[`e01`:`biolink:capable_of`]->(`n1`:`biolink:PhenotypicFeature`)" in clause

def test_invertible():
    """Test edge predicate inversion ca. biolink model 3.2.0."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:PhenotypicFeature",
            },
        },
        "edges": {
            "e01": {
                "predicates": "biolink:phenotype_of",
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    clause = get_query(qgraph, reasoner=False)
    # edges with types should be directed
    assert "(`n1`:`biolink:PhenotypicFeature`)-[`e01`:`biolink:has_phenotype`]->(`n0`:`biolink:Disease`)" in clause


def test_curie_int():
    """Test unusual curie formats."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
                "ids": 12,
            },
        },
        "edges": dict(),
    }
    clause = get_query(qgraph, reasoner=False)
    # the curie integer should be converted to a string
    assert "{`id`: \"12\"}" in clause



"""Test query arguments."""
import pytest
from reasoner_transpiler.cypher import get_query
from .fixtures import fixture_neo4j_driver


def test_skip_limit(neo4j_driver):
    """Test SKIP and LIMIT."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
                "ids": ["MONDO:0005148"],
            },
            "n1": {
                "categories": "biolink:ChemicalSubstance",
            },
        },
        "edges": {
            "e01": {
                "subject": "n1",
                "object": "n0",
                "predicates": "biolink:treats",
            },
        },
    }
    all_results = []
    with pytest.raises(NotImplementedError):
        output = neo4j_driver.run(get_query(qgraph, limit=2), convert_to_trapi=True, qgraph=qgraph)
    with pytest.raises(NotImplementedError):
        output = neo4j_driver.run(get_query(qgraph, skip=2), convert_to_trapi=True, qgraph=qgraph)

    """
    the following should pass if skip and limit are implemented
    output = neo4j_driver.run(get_query(qgraph, limit=2), convert_to_trapi=True, qgraph=qgraph)
    all_results.extend(output["results"])
    assert len(output["results"]) == 2
    output = neo4j_driver.run(get_query(qgraph, skip=2, limit=2), convert_to_trapi=True, qgraph=qgraph)
    all_results.extend(output["results"])
    assert len(output["results"]) == 1
    assert {
        "CHEBI:6801", "CHEBI:47612", "CHEBI:136043",
    } == set(
        result["node_bindings"]["n1"][0]["id"]
        for result in all_results
    )
    """


def test_max_connectivity(neo4j_driver):
    """Test max_connectivity option."""
    qgraph = {
        "nodes": {
            "n0": {
                "categories": "biolink:Disease",
            },
            "n1": {
                "categories": "biolink:ChemicalSubstance",
                "ids": ["CHEBI:6801"],
            },
        },
        "edges": {
            "e01": {
                "predicates": "biolink:treats",
                "subject": "n1",
                "object": "n0",
            },
        },
    }
    output = neo4j_driver.run(get_query(
        qgraph,
        max_connectivity=5,
    ), convert_to_trapi=True, qgraph=qgraph)
    assert len(output["results"]) == 2
    results = sorted(
        output["knowledge_graph"]["nodes"].values(),
        key=lambda node: node["name"],
    )
    expected_nodes = ["carcinoma", "metformin", "obesity disorder"]
    for ind, node in enumerate(results):
        assert node["name"] == expected_nodes[ind]


def test_use_hints():
    """Test unusual curie formats."""
    qgraph = {
        "nodes": {
            "n0": {
                "ids": [
                    ["NCBIGene:841"],
                ],
                "categories": "biolink:Gene",
            },
            "n1": {
            },
        },
        "edges": {
            "e01": {
                "predicates": [
                    "biolink:interacts_with",
                ],
                "subject": "n1",
                "object": "n0",
            },
        },
    }
    clause = get_query(qgraph, use_hints=True, reasoner=False)
    assert "USING INDEX" in clause

    clause = get_query(qgraph, use_hints=False, reasoner=False)
    assert "USING INDEX" not in clause



from .fixtures import fixture_neo4j_driver, fixture_async_neo4j_driver
from reasoner_transpiler.cypher import get_query
import asyncio


def test_bolt_driver_transform_results(neo4j_driver):
    qgraph = {
        "nodes": {
            "n0": {"ids": [
                "MONDO:0000001",
                "HP:0000118",
            ]},
            "n1": {},
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    output = neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph)
    assert len(output['results']) == 15
    for result in output["results"]:
        assert len(result["node_bindings"]) == 2
        assert len(result["analyses"]) == 1
    assert len(output['knowledge_graph']['nodes']) == 13
    assert len(output['auxiliary_graphs']) == 14


def test_bolt_async_driver_transform_results(async_neo4j_driver):
    qgraph = {
        "nodes": {
            "n0": {"ids": [
                "MONDO:0000001",
                "HP:0000118",
            ]},
            "n1": {},
        },
        "edges": {
            "e01": {
                "subject": "n0",
                "object": "n1",
            },
        },
    }
    output = asyncio.run(async_neo4j_driver.run(get_query(qgraph), convert_to_trapi=True, qgraph=qgraph))
    assert len(output['results']) == 15
    for result in output["results"]:
        assert len(result["node_bindings"]) == 2
        assert len(result["analyses"]) == 1
    assert len(output['knowledge_graph']['nodes']) == 13
    assert len(output['auxiliary_graphs']) == 14



import pytest
from ast import literal_eval
from reasoner_transpiler.matching import EdgeReference, set_predicates_in_graph, reset_predicates_in_graph
from reasoner_transpiler.exceptions import NoPossibleResultsException

def test_symmetric():
    """When a parent is symmetric, it can have both symmetric and directed children.   All of the symmetric and only the canonical directed children
    should appear in the label.   So in this example, only "biomarker_for" should appear, not "has_biomarker".
    There is no need to include the direction of the directed edge, because whichever is found will be consistant with the
    symmetric parent."""
    edge = {"subject": "s", "object": "o", "predicates": "biolink:correlated_with"}
    ref = EdgeReference("e0",edge,invert=True)
    preds = ref.label.split('|')
    assert len(preds) == 6
    assert "`biolink:biomarker_for`" in preds
    assert not ref.directed
    assert len(ref.filters) == 0
    assert not ref.cypher_invert

def test_directed_canonical():
    """For a directed parent predicate, all the children should be directed.  This should just have the
    subpredicates.  There's no WHERE block needed because all of the canonical versions of the predicates
    point the same direction by construction, and these are included in the --> syntax of the cypher.
    That happens in EdgeReference.str() if reference.directed is True."""
    edge= {"subject": "s", "object": "o", "predicates": "biolink:affects"}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    assert len(preds) == 10
    assert ref.directed
    assert len(ref.filters) == 0
    assert not ref.cypher_invert

def test_noncanonical():
    """If we send in a non-canonical (and by definition directed) query, we expect that we'll have a directed
    query with the reversed edge, the canonical predicate (and sub-predicates), no where clause."""
    edge= {"subject": "s", "object": "o", "predicates": "biolink:affected_by"}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    assert len(preds) == 10
    assert "`biolink:has_adverse_event`" in preds
    assert "`biolink:has_side_effect`" in preds
    assert ref.directed
    assert len(ref.filters) == 0
    #assert that the edge was reversed
    assert ref.cypher_invert

def test_multiple_canonical():
    """Make sure that the canonical logic is applied when there are multiple (canonical) predicates"""
    edge= {"subject": "s", "object": "o", "predicates": ["biolink:ameliorates_condition", "biolink:affects"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    #  affects, affects_response_to, ameliorates_condition, decreases_response_to, disrupts, exacerbates_condition,
    #  has_adverse_event, has_side_effect, increases_response_to, regulates
    assert len(preds) == 10
    assert ref.directed
    assert len(ref.filters) == 0
    assert not ref.cypher_invert

def test_multiple_noncanonical():
    """Make sure that the canonical logic is applied when there are multiple (noncanonical) predicates"""
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:condition_ameliorated_by",
                                                          "biolink:likelihood_affected_by"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    expected_preds = [ f"`biolink:{x}`" for x in ["ameliorates_condition", "predisposes_to_condition",
                                                  "preventative_for_condition", "affects_likelihood_of"]]
    assert set(preds) == set(expected_preds)  # order doesn't matter
    assert ref.directed
    assert len(ref.filters) == 0
    assert ref.cypher_invert

def parse_filter(filter):
    """A helper that takes the string filter and breaks it into a dict for comparison to expectations.
    Filters look like this:
    ['(type(`e0`) in ["biolink:treats", "biolink:ameliorates"] AND startNode(`e0`) = `s`) OR ...
    each OR creates a dict item relating the start node to its predicates:
    {"s": set(["biolink:treats", "biolink:ameliorates"])}
    """
    parsed_filter = {}
    clauses = filter.split(" OR ")
    for clause in clauses:
        subclauses = clause[1:-1].split(" AND ")
        for subclause in subclauses:
            if subclause.startswith("startNode"):
                startnode = subclause[-2] #s or o
            else:
                predicate_set = set( literal_eval(subclause.split(" in ")[-1]))
        parsed_filter[startnode] = predicate_set
    return parsed_filter

def test_multiple_conflicting():
    """Suppose that there are two predicates, one canonical, one not.  e.g. ameliorates_condition(canonical) and
    condition_predisposed_by(non canonical).   In this case, we expect a non-directed cypher edge with all canonical
    predicates, and a where clause separating out the ones going left to right from the ones going right to left."""
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:ameliorates_condition",
                                                          "biolink:likelihood_affected_by"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    #Make sure that the label contains only canonical predicates
    expected_preds = [ f"`biolink:{x}`" for x in ["ameliorates_condition", "predisposes_to_condition",
                                                  "preventative_for_condition", "affects_likelihood_of"]]
    assert set(preds) == set(expected_preds) #order doesn't matter
    assert not ref.directed
    #filters should look like:
    #['(type(`e0`) in ["biolink:treats", "biolink:ameliorates"] AND startNode(`e0`) = `s`) OR (type(`e0`) in ["biolink:predisposes", "biolink:prevents", "biolink:affects_risk_for"] AND startNode(`e0`) = `o`)']
    expected_filter = {"s": set(["biolink:ameliorates_condition"]),
                       "o": set(["biolink:predisposes_to_condition",
                                 "biolink:preventative_for_condition",
                                 "biolink:affects_likelihood_of"])}
    assert len(ref.filters) == 1
    parsed_filter = parse_filter(ref.filters[0])
    assert parsed_filter == expected_filter
    assert not ref.cypher_invert

def test_symmetric_canonical():
    """Two predicates.  One symmetric, one canonical/directed.   We would expect a non-directed, non-inverted cypher
    And a where clause.   the canonical should be in one of the subclauses, and all subclasses of the symmetric
    (including any canonical/directed subclasses) should be in both."""
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:correlated_with", "biolink:affects_likelihood_of"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    # Make sure that the label contains only canonical predicates
    correlated_sub_preds = ["correlated_with", "biomarker_for", "coexpressed_with", "negatively_correlated_with",
                            "positively_correlated_with", "occurs_together_in_literature_with"]
    likelihood_of_sub_preds = ["predisposes_to_condition", "preventative_for_condition", "affects_likelihood_of"]
    expected_preds = [f"`biolink:{x}`" for x in correlated_sub_preds + likelihood_of_sub_preds]
    assert set(preds) == set(expected_preds)  # order doesn't matter
    assert not ref.directed
    expected_filter = {"s": set([f"biolink:{x}" for x in correlated_sub_preds + likelihood_of_sub_preds]),
                       "o": set([f"biolink:{x}" for x in correlated_sub_preds])}
    assert len(ref.filters) == 1
    parsed_filter = parse_filter(ref.filters[0])
    assert parsed_filter == expected_filter
    assert not ref.cypher_invert

def test_symmetric_noncanonical():
    """Two predicates.  One symmetric, one noncanonical/directed.   We would expect a non-directed, non-inverted cypher
    And a where clause.   the canonical should be in one of the subclauses, and all subclasses of the symmetric
    (including any canonical/directed subclasses) should be in both. Note that the ameliorates subpredicates
    have the object as the starting node in this case, unlike the canonical case above"""
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:correlated_with", "biolink:likelihood_affected_by"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    # Make sure that the label contains only canonical predicates
    correlated_sub_preds = ["correlated_with", "biomarker_for", "coexpressed_with", "negatively_correlated_with",
                            "positively_correlated_with", "occurs_together_in_literature_with"]
    ameliorates_sub_preds = ["predisposes_to_condition", "preventative_for_condition", "affects_likelihood_of"]
    expected_preds = [f"`biolink:{x}`" for x in correlated_sub_preds + ameliorates_sub_preds]
    assert set(preds) == set(expected_preds)  # order doesn't matter
    assert not ref.directed
    expected_filter = {"o": set([f"biolink:{x}" for x in correlated_sub_preds + ameliorates_sub_preds]),
                       "s": set([f"biolink:{x}" for x in correlated_sub_preds])}
    assert len(ref.filters) == 1
    parsed_filter = parse_filter(ref.filters[0])
    assert parsed_filter == expected_filter
    assert not ref.cypher_invert

def test_related_to():
    """If the predicate is the top-level predicate, we don't really need any subpredicates etc.  We'll accept any
    edge."""
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:related_to"]}
    ref = EdgeReference("e0", edge, invert=True)
    assert ref.label == ""
    assert not ref.directed
    assert not ref.cypher_invert
    assert len(ref.filters) == 0

def test_no_predicate():
    """If the predicate is missing, we'll accept any edge."""
    edge = {"subject": "s", "object": "o"}
    ref = EdgeReference("e0", edge, invert=True)
    assert ref.label == ""
    assert not ref.directed
    assert not ref.cypher_invert
    assert len(ref.filters) == 0


def test_predicates_in_graph_setting():
    """Make sure that predicates aren't included if they aren't in the graph"""
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:treats_or_applied_or_studied_to_treat"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    expected_preds = [f"`biolink:{x}`" for x in ["ameliorates_condition", "applied_to_treat",
                                                 "beneficial_in_models_for", "in_clinical_trials_for",
                                                 "in_preclinical_trials_for",
                                                 "preventative_for_condition", "studied_to_treat",
                                                 "treats", "treats_or_applied_or_studied_to_treat"]]
    assert set(preds) == set(expected_preds)

    set_predicates_in_graph(["biolink:ameliorates_condition", "biolink:applied_to_treat",
                             "biolink:beneficial_in_models_for", "biolink:in_clinical_trials_for",
                             "biolink:in_preclinical_trials_for"])

    """Make sure that predicates aren't included if they aren't in the graph"""
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:treats_or_applied_or_studied_to_treat"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    expected_preds = [f"`biolink:{x}`" for x in ["ameliorates_condition", "applied_to_treat",
                                                 "beneficial_in_models_for", "in_clinical_trials_for",
                                                 "in_preclinical_trials_for"]]
    assert set(preds) == set(expected_preds)
    reset_predicates_in_graph()


def test_predicates_in_graph_setting_no_results():

    set_predicates_in_graph(["biolink:affects", "biolink:is_nonsense_variant_of"])

    # related_to should get turned into no predicates but shouldn't throw the NoPossibleResultsException
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:related_to"]}
    ref = EdgeReference("e0", edge, invert=True)
    preds = ref.label.split('|')
    expected_preds = ['']
    assert preds == expected_preds

    # treats_or_applied_or_studied_to_treat nor it's descendants are in the graph, throw NoPossibleResultsException
    edge = {"subject": "s", "object": "o", "predicates": ["biolink:treats_or_applied_or_studied_to_treat"]}
    with pytest.raises(NoPossibleResultsException):
        ref = EdgeReference("e0", edge, invert=True)
    reset_predicates_in_graph()



"""Tools for compiling QGraph into Cypher query."""
import os
import json

from collections import defaultdict
from neo4j import AsyncResult, Result

from .attributes import transform_attributes, EDGE_SOURCE_PROPS
from .matching import match_query


PROVENANCE_TAG = os.environ.get('PROVENANCE_TAG', 'reasoner-transpiler')


def nest_op(operator, *args):
    """Generate a nested set of operations from a flat expression."""
    if len(args) > 2:
        return [operator, args[0], nest_op(operator, *args[1:])]
    else:
        return [operator, *args]


def assemble_results(qnodes, qedges, **kwargs):
    """Assemble results into Reasoner format."""
    clauses = []

    for qnode in qnodes.values():
        if qnode.get("set_interpretation", "") == "MANY":
            raise NotImplementedError(f'This feature is currently not implemented: set_interpretation=MANY')

    # TODO - implement pagination (SKIP and LIMIT)
    #  right now SKIP and LIMIT are unsupported and this will throw NotImplementedError if they are requested
    pagination(**kwargs)

    nodes = [f"`{qnode_id}`.id" for qnode_id, qnode in qnodes.items()]
    edges = [f"elementId(`{qedge_id}`)" if not qedge.get('_subclass', False) else f"[x in `{qedge_id}` | elementId(x)]"
             for qedge_id, qedge in qedges.items()]
    if nodes or edges:
        nodes_assemble = " + ".join([
            f"collect(`{qnode_id}`)"
            for qnode_id, qnode in qnodes.items()
        ])
        if not nodes_assemble:
            nodes_assemble = '[]'
        edges_assemble = " + ".join([
            f"collect([elementId(`{qedge_id}`), startNode(`{qedge_id}`).id, "
            f"type(`{qedge_id}`), endNode(`{qedge_id}`).id, properties(`{qedge_id}`)])"
            if not qedge.get('_subclass', False) else
            f"collect([x in `{qedge_id}` | [elementId(x), startNode(x).id, "
            f"type(x), endNode(x).id, properties(x)]])"
            for qedge_id, qedge in qedges.items()
        ])
        if not edges_assemble:
            edges_assemble = '[]'
        assemble_clause = f"WITH apoc.coll.toSet({nodes_assemble}) AS nodes, apoc.coll.toSet({edges_assemble}) AS edges, collect(DISTINCT ["

        if nodes:
            assemble_clause += ', '.join(nodes)
            if edges:
                assemble_clause += ', '
        if edges:
            assemble_clause += ', '.join(edges)
        assemble_clause += "]) AS paths "
        clauses.append(assemble_clause)
        return_clause = "RETURN nodes, edges, paths"
    else:
        return_clause = 'RETURN [] as nodes, [] as edges, [] as paths'

    clauses.append(return_clause)
    return clauses


def pagination(skip=None, limit=None, **kwargs):
    """Get pagination clauses."""
    clauses = []
    if skip is not None:
        raise NotImplementedError(f'SKIP is not currently implemented.')
        # clauses.append(f"SKIP {skip}")
    if limit is not None:
        raise NotImplementedError(f'LIMIT is not currently implemented.')
        # clauses.append(f"LIMIT {limit}")
    return clauses


def get_query(qgraph, **kwargs):
    """Generate a Cypher query to extract the answer maps for a question.

    Returns the query as a string.
    """
    # commented this out because now we rely on the altering the qgraph to transform results into TRAPI,
    # leaving as a reminder in case that breaks something
    # qgraph = copy.deepcopy(qgraph)
    clauses = []
    query = match_query(qgraph, **kwargs)
    clauses.extend(query.compile())
    where_clause = query.where_clause()
    if where_clause:
        if not clauses[-1].startswith("WITH"):
            clauses.append(query.with_clause())
        clauses.append(where_clause)

    if not kwargs.pop("reasoner", True):
        clauses.append(query.return_clause())
        # add SKIP and LIMIT sub-clauses
        clauses.extend(pagination(**kwargs))
    else:
        clauses.extend(assemble_results(
            query.qgraph["nodes"],
            query.qgraph["edges"],
            **kwargs,
        ))

    return " ".join(clauses)


def transform_result(cypher_record,
                     qgraph: dict):

    nodes, edges, paths = unpack_bolt_record(cypher_record)

    # Convert the list of unique result nodes from cypher results to dictionaries
    # then convert them to TRAPI format, constructing the knowledge_graph["nodes"] section of the TRAPI response
    kg_nodes = transform_nodes_list(nodes)

    # Convert the list of unique edges from cypher results to dictionaries
    # then convert them to TRAPI format, constructing the knowledge_graph["edges"] section of the TRAPI response.
    # Also make a mapping of the neo4j element_id to the edge id to be used in the TRAPI edge bindings
    # the edge id used in TRAPI is the 'id' property on the edge if there is one, otherwise assigned integers 0,1,2..
    kg_edges, element_id_to_edge_id = transform_edges_list(edges)

    results = {}  # results are grouped by unique sets of result node ids
    aux_graphs = {}  # auxiliary_graphs

    qgraph_nodes = qgraph["nodes"]
    qnodes_with_set_interpretation_all = {qnode_id for qnode_id, qnode in qgraph_nodes.items()
                                          if qnode.get('set_interpretation', 'BATCH') == 'ALL'}
    qnodes_with_superclass_nodes = {qnode_id for qnode_id, qnode in qgraph_nodes.items()
                                    if f'{qnode_id}_superclass' in qgraph_nodes}

    qgraph_edges = qgraph["edges"]
    qedges_with_attached_subclass_edges = defaultdict(list)
    for qedge_id, qedge in qgraph_edges.items():
        if not qedge.get('_subclass', False):
            if qedge["subject"] in qnodes_with_superclass_nodes:
                qedges_with_attached_subclass_edges[qedge_id].append(
                    ('subject', f'{qedge["subject"]}_subclass_edge', f'{qedge["subject"]}_superclass'))
            if qedge["object"] in qnodes_with_superclass_nodes:
                qedges_with_attached_subclass_edges[qedge_id].append(
                    ('object', f'{qedge["object"]}_subclass_edge',  f'{qedge["object"]}_superclass'))

    # Each path is an array of nodes and edges like [n1, n2, n3, e1, e2, e3],
    # where nodes are node_ids from the graph and edges are element_ids of relationships from the graph.
    for path in paths:

        # Map results/paths to their corresponding qnodes and qedges
        qnode_id_to_results = {qnode_id: (qnode, result_node_id) for (qnode_id, qnode), result_node_id in
                               zip(qgraph_nodes.items(), path[:len(qgraph_nodes)])}
        qedge_id_to_results = {qedge_id: (qedge, result_edge_id) for (qedge_id, qedge), result_edge_id in
                               zip(qgraph_edges.items(), path[-len(qgraph_edges):])}

        # results are grouped by unique sets of nodes, concatenating node ids to this string generates the key
        result_key = ""

        # create TRAPI node bindings
        edge_bindings = {}
        node_bindings = {}
        for qnode_id, (qnode, result_node_id) in qnode_id_to_results.items():

            # don't return superclass qnodes in the node bindings
            if qnode.get('_superclass', False):
                continue
            # if there isn't a result for this node set an empty node binding
            if not result_node_id:
                node_bindings[qnode_id] = []
                continue

            # create a node binding
            if qnode_id in qnodes_with_set_interpretation_all:
                # if qnode has set_interpretation=ALL there won't be any superclass bindings
                node_bindings[qnode_id] = [{'id': result_node_id, 'attributes': []}]

            elif qnode_id in qnodes_with_superclass_nodes:
                # If the qnode has a superclass node, and it has a different result node id,
                # use the superclass result node id instead, because it's the actual query id.
                # We used to include the query_id property here to show that, but now we're making a support graph
                # that can represent the underlying subclass edge(s).
                superclass_qnode, superclass_result_id = qnode_id_to_results[f'{qnode_id}_superclass']
                node_bindings[qnode_id] = \
                    [{'id': result_node_id if superclass_result_id == result_node_id else superclass_result_id,
                      'attributes': []}]
            else:
                # Otherwise, create a normal node binding.
                node_bindings[qnode_id] = \
                    [{'id': result_node_id,
                      'attributes': []}]

            # add the result node id to the result key
            result_key += node_bindings[qnode_id][0]['id']

        # Create TRAPI edge bindings
        for qedge_id, (qedge, path_edge) in qedge_id_to_results.items():

            # skip empty results
            if not path_edge:
                continue
            # don't return subclass qedges in the edge bindings
            if qedge.get("_subclass", False):
                continue

            # find the knowledge graph edge id for the element id from the path edge
            edge_element_id = path_edge
            graph_edge_id = element_id_to_edge_id[edge_element_id]

            # Check to see if the edge has subclass edges that are connected to it
            subclass_edge_ids = []
            superclass_node_ids = {}
            for (subclass_subject_or_object, subclass_qedge_id, superclass_qnode_id) in qedges_with_attached_subclass_edges.get(qedge_id, []):
                # If so, check to see if there are results for it
                qedge, subclass_edge_element_ids = qedge_id_to_results[subclass_qedge_id]
                if subclass_edge_element_ids:
                    # If path_edge is Truthy, it means the subclass was used in the result.
                    # For subclass edges, path result is a list of element ids, due to being a variable length edge.
                    # make a list of the subclass edges plus the result edge from the query.
                    subclass_edge_ids.extend([element_id_to_edge_id[ele_id] for ele_id in subclass_edge_element_ids])

                    qnode, superclass_result_node_id = qnode_id_to_results[superclass_qnode_id]
                    superclass_node_ids[subclass_subject_or_object] = superclass_result_node_id

            if subclass_edge_ids:
                # make a composite id with all of their kg edge ids
                composite_edge_ids = [graph_edge_id] + subclass_edge_ids
                composite_edge_id = "_".join(composite_edge_ids)
                aux_graph_id = f"aux_{composite_edge_id}"
                if aux_graph_id not in aux_graphs:
                    aux_graphs[aux_graph_id] = {
                        "edges": composite_edge_ids,
                        "attributes": []
                    }
                if composite_edge_id not in kg_edges:
                    real_edge = kg_edges[graph_edge_id]
                    inferred_result_edge = {"subject": real_edge["subject"],
                                            "predicate": real_edge["predicate"],
                                            "object": real_edge["object"],
                                            "attributes": [
                                                {
                                                    "attribute_type_id": "biolink:knowledge_level",
                                                    "value": "logical_entailment"
                                                },
                                                {
                                                    "attribute_type_id": "biolink:agent_type",
                                                    "value": "automated_agent",
                                                },
                                                {
                                                    "attribute_type_id": "biolink:support_graphs",
                                                    "value": [aux_graph_id]
                                                }
                                            ],
                                            "sources": [
                                                {
                                                    "resource_id": PROVENANCE_TAG,
                                                    "resource_role": "primary_knowledge_source"
                                                }
                                            ],
                                            # this overwrites subject and/or object with the superclass node ids
                                            # if they exist
                                            **superclass_node_ids}
                    kg_edges[composite_edge_id] = inferred_result_edge

                # make an edge binding with the inferred subclass edge
                edge_bindings[qedge_id] = [{'id': composite_edge_id, 'attributes': []}]
            else:
                # if no subclass edges for this edge make a normal edge binding
                edge_bindings[qedge_id] = [{'id': graph_edge_id, 'attributes': []}]

        if result_key != '':  # avoid adding results for the default node binding key ''
            # if we haven't encountered this specific group of result nodes before, create a new result
            if result_key not in results:
                results[result_key] = {'node_bindings': node_bindings,
                                       'analyses': [{'edge_bindings': edge_bindings,
                                                     'resource_id': PROVENANCE_TAG}]}
            else:
                # otherwise append new edge bindings to the existing result
                for qedge_id, edge_binding_list in edge_bindings.items():
                    results[result_key]['analyses'][0]['edge_bindings'][qedge_id].extend(
                        [new_edge_bind for new_edge_bind in edge_binding_list if new_edge_bind['id'] not in
                         [existing_edge_bind['id'] for existing_edge_bind in
                          results[result_key]['analyses'][0]['edge_bindings'][qedge_id]]])

    knowledge_graph = {
        'nodes': kg_nodes,
        'edges': kg_edges
    }
    transformed_results = {
        'results': list(results.values()),  # convert the results dictionary to a flattened list
        'knowledge_graph': knowledge_graph,
        'auxiliary_graphs': aux_graphs
    }
    return transformed_results


def transform_nodes_list(nodes):
    kg_nodes = {}
    for cypher_node in nodes:
        node = convert_bolt_node_to_dict(cypher_node)
        kg_nodes[node['id']] = {
            'name': node['name'],
            'categories': sorted(node.pop('labels')),
            **transform_attributes(node, node=True)}
    return kg_nodes


def transform_edges_list(edges):
    kg_edges = {}
    element_id_to_edge_id = {}
    for edge_index, cypher_edge_result in enumerate(edges):
        # skip an empty list
        if len(cypher_edge_result) == 0:
            cypher_edges = []
        # check to see if it's a list of lists (multiple edges)
        elif isinstance(cypher_edge_result[0], list):
            cypher_edges = cypher_edge_result
        else:
            # otherwise it's just one list (one edge)
            cypher_edges = list()  # this looks weird, but it's necessary to make a list of one list (I think?)
            cypher_edges.append(cypher_edge_result)

        # transform the edge into TRAPI
        for cypher_edge in cypher_edges:
            edge_element_id, edge_dict = convert_bolt_edge_to_dict(cypher_edge)
            edge_id = edge_dict.get('id', edge_index)
            element_id_to_edge_id[edge_element_id] = edge_id
            # get properties matching EDGE_SOURCE_PROPS keys, remove biolink: if needed,
            # then pass (key, value) tuples to construct_sources_tree for formatting, constructing the 'sources' section
            edge_dict['sources'] = construct_sources_tree([
                    (edge_source_prop.removeprefix('biolink:'), edge_dict.get(edge_source_prop))
                    for edge_source_prop in EDGE_SOURCE_PROPS if edge_dict.get(edge_source_prop, None)])
            # convert all remaining attributes to TRAPI format
            edge_dict.update(transform_attributes(edge_dict, node=False))
            kg_edges[edge_id] = edge_dict
    return kg_edges, element_id_to_edge_id


# This function takes EDGE_SOURCE_PROPS properties from results, converts them into proper
# TRAPI dictionaries, and assigns the proper upstream ids to each resource. It does not currently attempt to avoid
# duplicate aggregator results, which shouldn't exist in the graphs.
def construct_sources_tree(sources):

    # first find the primary knowledge source, there should always be one
    primary_knowledge_source = None
    formatted_sources = None
    for resource_role, resource_id in sources:
        if resource_role == "primary_knowledge_source":
            primary_knowledge_source = resource_id
            # add it to the formatted TRAPI output
            formatted_sources = [{
                "resource_id": primary_knowledge_source,
                "resource_role": "primary_knowledge_source"
            }]
    if not primary_knowledge_source:
        # we could hard fail here, every edge should have a primary ks, but I haven't fixed all the tests yet
        #     raise KeyError(f'primary_knowledge_source missing from sources section of cypher results! '
        #                    f'sources: {sources}')
        return []

    # then find any aggregator lists
    aggregator_list_sources = []
    for resource_role, resource_id in sources:
        # this looks weird but the idea is that you could have a few parallel lists like:
        # aggregator_knowledge_source, aggregator_knowledge_source_2, aggregator_knowledge_source_3
        if resource_role.startswith("aggregator_knowledge_source"):
            aggregator_list_sources.append(resource_id)
    # walk through the aggregator lists and construct the chains of provenance
    terminal_aggregators = set()
    for aggregator_list in aggregator_list_sources:
        # each aggregator list should be in order, so we can deduce the upstream chains
        last_aggregator = None
        for aggregator_knowledge_source in aggregator_list:
            formatted_sources.append({
                "resource_id": aggregator_knowledge_source,
                "resource_role": "aggregator_knowledge_source",
                "upstream_resource_ids": [last_aggregator] if last_aggregator else [primary_knowledge_source]
            })
            last_aggregator = aggregator_knowledge_source
        # store the last aggregator in the list, because this will be an upstream source for the plater one
        terminal_aggregators.add(last_aggregator)
    # add PROVENANCE_TAG as the most downstream aggregator,
    # it will have as upstream either the primary ks or all of the furthest downstream aggregators if they exist
    # this will be used by applications like Plater which need to append themselves as an aggregator
    formatted_sources.append({
        "resource_id": PROVENANCE_TAG,
        "resource_role": "aggregator_knowledge_source",
        "upstream_resource_ids": list(terminal_aggregators) if terminal_aggregators else [primary_knowledge_source]
    })
    return list(formatted_sources)


def convert_bolt_node_to_dict(bolt_node):
    if not bolt_node:
        return None
    node = {key: value for key, value in bolt_node.items()}
    # node['element_id'] = bolt_node.element_id
    node['labels'] = bolt_node.labels
    return node


def convert_jolt_node_to_dict(jolt_node):
    if not jolt_node:
        return None
    node = jolt_node['()'][2]
    node['element_id'] = jolt_node['()'][0]
    node['labels'] = jolt_node['()'][1]
    return node


def convert_bolt_edge_to_dict(bolt_edge):
    if not bolt_edge:
        print(f'Tried to convert a missing edge: {bolt_edge}')
        return None, None

    # Convert a list representing an edge from cypher results into a dictionary.
    # This is not a standard Edge object from the bolt driver, it's a list product of a specific cypher return format.
    # This is done to prevent including redundant node information on every edge.
    # See the cypher generated in the edges_assemble clause in assemble_results() for more details.
    element_id = bolt_edge[0]
    converted_edge = {
        'subject': bolt_edge[1],
        'predicate': bolt_edge[2],
        'object': bolt_edge[3],
        **bolt_edge[4]
    }
    return element_id, converted_edge


def convert_jolt_edge_to_dict(jolt_edges, jolt_element_id_lookup):
    if not jolt_edges:
        return None
    convert_edges = []
    jolt_edges = [jolt_edges] if not isinstance(jolt_edges, list) else jolt_edges
    for jolt_edge in jolt_edges:
        jolt_edge = next(iter(jolt_edge.values()))
        edge = jolt_edge[4]  # start with the properties
        # add the SPO from the other items of the jolt edge list
        edge.update({
            'subject': jolt_element_id_lookup[jolt_edge[1]],
            'predicate': jolt_edge[2],
            'object': jolt_element_id_lookup[jolt_edge[3]]})
        if 'id' not in edge:
            edge['id'] = jolt_edge[0]  # if the edge didn't have an "id" use the neo4j element id
        convert_edges.append(edge)
    return convert_edges


def unpack_bolt_record(bolt_record):
    return bolt_record['nodes'], bolt_record['edges'], bolt_record['paths']


def unpack_jolt_result(jolt_response):
    headers = []
    for line in jolt_response.split("\n"):
        line = json.loads(line)
        if 'header' in line:
            headers = line['header']['fields']
        elif 'data' in line:
            data = {header: data_item for (header, data_item) in zip(headers, line['data'])}
            return data['nodes'], data['edges'], data['paths']



"""MATCHing tools."""
from typing import Dict, List

from .biolink import bmt, ALL_BIOLINK_ENUMS
from .exceptions import InvalidPredicateError, InvalidQualifierError, InvalidQualifierValueError, UnsupportedError, NoPossibleResultsException
from .nesting import Query
from .util import ensure_list, snake_case, space_case, pascal_case


# A placeholder for an optional set of predicates to be used as a filter when constructing cypher queries.
# If not None/empty, only the predicates in this list will be included in cypher queries.
PREDICATES_IN_GRAPH = None


def set_predicates_in_graph(predicates: set):
    global PREDICATES_IN_GRAPH
    if predicates:
        PREDICATES_IN_GRAPH = {space_case(p.removeprefix('biolink:')) for p in predicates}


def reset_predicates_in_graph():
    global PREDICATES_IN_GRAPH
    PREDICATES_IN_GRAPH = None


def cypher_prop_string(value):
    """Convert property value to cypher string representation."""
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, str):
        return "\"{0}\"".format(
            value.replace("\"", "\\\"")
        )
    if isinstance(value, (float, int)):
        return str(value)
    raise UnsupportedError(f"Unsupported property type: {type(value).__name__}.")


def convert_constraints(constraints):
    props = {}
    for constraint in constraints:
        try:
            constraint["id"] = constraint["id"].removeprefix("biolink:")
            operator = constraint.get("operator", "===")
            if operator == "===":
                props[constraint["id"]] = constraint["value"]
            elif operator == "==":
                if isinstance(constraint["value"], list):
                    # TODO we should be able to support this but it might need to go in the where clause
                    # operator is == and value is a list [1,2,3] means return results where value = 1, 2, OR 3
                    raise UnsupportedError(f'Unsupported attribute constraint: {constraint}')
                else:
                    props[constraint["id"]] = constraint["value"]
        except KeyError:
            raise UnsupportedError(f'Invalid attribute constraint: {constraint}')
    return props


class NodeReference:
    """Node reference object."""

    def __init__(self, node_id, node, **kwargs):
        """Create a node reference.

        All node properties of types [str, bool, float/int] will be enforced
        verbatim EXCEPT for the following reserved properties:
        * categories
        * ids
        * name
        * set_interpretation
        Un-reserved properties with other types will be coerced to str.
        """
        max_connectivity = kwargs.get("max_connectivity", -1)
        self.anonymous = kwargs.get("anonymous", False)

        node = dict(node)  # shallow copy
        self.name = f"`{node_id}`"

        props = {}
        self._filters = []
        self.labels = []

        category = node.pop("categories", ["biolink:NamedThing"])
        if isinstance(category, list) and len(category) == 1:
            category = category[0]
        elif category is None:
            category = "biolink:NamedThing"
        if isinstance(category, list):
            self.labels = ['biolink:NamedThing']
            self._filters.append(" OR ".join([
                "{1} in labels({0})".format(
                    self.name,
                    cypher_prop_string(ci)
                )
                for ci in category
            ]))
        elif category is not None:
            # coerce to a string
            self.labels = [str(category)]

        curie = node.pop("ids", None)
        if isinstance(curie, list) and len(curie) == 1:
            curie = curie[0]
        if isinstance(curie, list):
            self._filters.append("{0}.id in [{1}]".format(
                self.name,
                ", ".join([
                    cypher_prop_string(ci)
                    for ci in curie
                ])
            ))
        elif curie is not None:
            # coerce to a string
            props["id"] = str(curie)

        if max_connectivity > -1:
            self._filters.append("COUNT {{ ({0})-[]-() }} < {1} + 1".format(
                self.name,
                max_connectivity,
            ))

        # the transpiler used to do the following,
        # but now TRAPI QNodes should only have attributes constraints in the "constraints" field
        # other properties should be not considered attributes to query on anymore
        # props.update(
        #     (key, value)
        #     for key, value in node.items()
        #     if key not in ("name", "set_interpretation", "constraints")
        # )

        props.update(convert_constraints(node.pop("constraints", [])))

        self.prop_string = "{" + ", ".join([
            f"`{key}`: {cypher_prop_string(value)}"
            for key, value in props.items()
            if value is not None and not key.startswith("_")
        ]) + "}" if props else ""
        self._hints = []
        if curie and self.labels:
            self._hints.append(f"USING INDEX {self.name}:`{self.labels[0]}`(id)")
        self._num = 0

    def __str__(self):
        """Return the cypher node reference."""
        self._num += 1
        if self._num > 1:
            return f"({self.name})"
        ref = "("
        elements = []
        if not self.anonymous:
            elements.append(self.name)
        if self.labels:
            elements.append("".join(f":`{label}`" for label in self.labels))
        elements = ["".join(elements)]
        if self.prop_string:
            elements.append(self.prop_string)
        ref = "({})".format(" ".join(elements))
        return ref

    @property
    def filters(self):
        """Return filters for the cypher node reference.

        To be used in a WHERE clause following the MATCH clause.
        """
        if self._num > 1:
            return []
        return self._filters

    @property
    def hints(self):
        """Return hints for the cypher node reference.

        To be appended to the MATCH clause.
        """
        if self._num > 1:
            return []
        return self._hints


class MissingReference(NodeReference):
    """Missing node reference object."""

    def __init__(self, name):  # pylint: disable=super-init-not-called
        """Initialize."""
        self.name = f"`{name}`"
        self._num = 2


class EdgeReference:
    """Edge reference object."""

    def __init__(
        self,
        edge_id,
        edge,
        anonymous=False,
        invert=True,
        **kwargs
    ):
        """Create an edge reference."""
        edge = dict(edge)  # make shallow copy
        _subject = edge.pop("subject")
        _object = edge.pop("object")
        self.name = edge_id if not anonymous else ""
        self.predicates: List[str] = ensure_list(
            edge.pop("predicates", []) or []
        )

        # "related_to" has every predicate as a descendent, it's inclusion is equivalent to no/any predicate
        if 'biolink:related_to' in self.predicates:
            self.predicates = []
        queried_predicates = self.predicates.copy()

        self.filters = []
        self.qualifier_filters = ""
        self.label = None  #What goes in the [] on the edge in cypher
        self.length = edge.pop("_length", (1, 1))
        invert = invert and edge.pop("_invert", True)

        self.inverse_predicates = []
        self.directed = False # Controls whether there is an arrow on the edge in cypher
        self.symmetric = True # Whether the original top-level predicates are all symmetric
        self.cypher_invert = False # If true, then the cypher source node will be subject, if false then object
        for predicate in self.predicates:
            el = bmt.get_element(space_case(predicate[8:]))
            if el:
                inverse_predicate = el.inverse
                if inverse_predicate is not None:
                    self.inverse_predicates.append(f"biolink:{snake_case(inverse_predicate)}")

                # if symmetric add to inverse list so we query in both directions
                if el.symmetric:
                    self.inverse_predicates.append(predicate)
                else:
                    self.symmetric = False
            else:
                error_message = f"Invalid predicate error: (predicate: {predicate}) is not " \
                                f"a valid biolink model predicate."
                raise InvalidPredicateError(error_message=error_message)

        # relationship is directed if any provided predicate is asymmetrical
        if not self.symmetric:
            self.directed = True

        # get all canonical and/or symmetric descendant predicates
        self.predicates = [
            f"biolink:{snake_case(p)}"
            for predicate in self.predicates
            for p in bmt.get_descendants(space_case(predicate[8:]))
            if ((not PREDICATES_IN_GRAPH) or p in PREDICATES_IN_GRAPH) and
            (bmt.get_element(p).annotations.get('canonical_predicate', False)
             or ('symmetric' in bmt.get_element(p) and bmt.get_element(p).symmetric))
        ]
        # get all canonical and/or symmetric descendant predicates of inverse predicates if invert flag is true
        # otherwise get rid of all inverse predicates
        self.inverse_predicates = [
            f"biolink:{snake_case(p)}"
            for predicate in self.inverse_predicates
            for p in bmt.get_descendants(space_case(predicate[8:]))
            if ((not PREDICATES_IN_GRAPH) or p in PREDICATES_IN_GRAPH) and
            (bmt.get_element(p).annotations.get('canonical_predicate', False)
             or ('symmetric' in bmt.get_element(p) and bmt.get_element(p).symmetric))
        ] if invert else []

        unique_preds = list(set(self.predicates + self.inverse_predicates))
        if queried_predicates and not unique_preds:
            raise NoPossibleResultsException(f'A query was made with the following predicates, '
                                             f'but none of them or their descendants are in the graph queried: '
                                             f'{queried_predicates}')
        #Having the predicates sorted doesn't matter to neo4j, but it helps in testing b/c we get a consistent string.
        unique_preds.sort()
        self.label = "|".join(
            f"`{predicate}`"
            for predicate in unique_preds
        )

        # We only need the WHERE clause if: we have canonical edges pointing in opposite directions.  In that
        # case we need a non-directed edge and a where clause that points the right ones in different directions
        if not self.symmetric:
            if self.inverse_predicates and self.predicates:
                self.directed = False
                self.filters.append(" OR ".join([
                    "(type(`{0}`) in [{1}] AND startNode(`{0}`) = `{2}`)".format(
                        self.name,
                        ", ".join([
                            f"\"{predicate}\""
                            for predicate in self.predicates
                        ]),
                        _subject,
                    ),
                    "(type(`{0}`) in [{1}] AND startNode(`{0}`) = `{2}`)".format(
                        self.name,
                        ", ".join([
                            f"\"{predicate}\""
                            for predicate in self.inverse_predicates
                        ]),
                        _object,
                    ),
                ]))
            #If we have inverse_predicates and not self.predicates then it means that everything in predicates
            # was non-canonical and we reversed them all so we need to invert the edge
            elif self.inverse_predicates:
                self.cypher_invert = True

        self.qualifier_filters = self.__qualifier_filters(edge, edge_id)

        props = convert_constraints(edge.pop("attribute_constraints", []))
        self.prop_string = " {" + ", ".join([
            f"`{key}`: {cypher_prop_string(value)}"
            for key, value in props.items()
            if value is not None
        ]) + "}" if props else ""

    def __qualifier_filters(self, edge, edge_id):
        constraints = edge.get("qualifier_constraints", [])
        ors = []
        for constraint in constraints:
            ands = []
            for constraint_filter in constraint.get("qualifier_set", []):
                # also handle "qualifier_set": [{ }]
                if not constraint_filter:
                    continue

                # remove the biolink prefix if there - expect graphs to not have curie prefixes on any edge properties
                qualifier_type = constraint_filter['qualifier_type_id'].removeprefix('biolink:')
                queried_qualifier_value = constraint_filter['qualifier_value']

                if not bmt.is_qualifier(qualifier_type):
                    raise InvalidQualifierError(f'Invalid qualifier in query: {qualifier_type}')

                # we should do something like this, it does not work without knowing the association type of the edge
                # if not bmt.validate_qualifier(qualifier_type_id=qualifier_type,
                #                               qualifier_value=queried_qualifier_value,
                #                               associations=[???]):
                #     raise InvalidQualifierError(f'Invalid qualifier requested,
                #     {qualifier_type}:{queried_qualifier_value}')

                qualifier_value_plus_descendants = [queried_qualifier_value]
                # qualified_predicate doesn't have an enum as values so the following does not apply
                if qualifier_type != 'qualified_predicate':
                    permissible_value = False
                    for enum_for_qualifier_values in ALL_BIOLINK_ENUMS:
                        if bmt.is_permissible_value_of_enum(enum_name=enum_for_qualifier_values,
                                                            value=queried_qualifier_value):
                            permissible_value = True
                            qualifier_value_plus_descendants += bmt.get_permissible_value_descendants(
                                permissible_value=queried_qualifier_value,
                                enum_name=enum_for_qualifier_values)
                    if not permissible_value:
                        raise InvalidQualifierValueError(
                            f'Invalid value for qualifier {qualifier_type} in query: {queried_qualifier_value}')

                # Join qualifier value hierarchy with an or
                qualifier_where_condition = " ( " + " OR ".join(
                    [f"`{edge_id}`.{qualifier_type} = {cypher_prop_string(qualifier_value)}" for qualifier_value in
                     set(qualifier_value_plus_descendants)]) + " ) "
                ands.append(qualifier_where_condition)
            # if qualifier set is empty ; loop to the next
            if not len(ands):
                continue
            # join contraints in a single qualifier set with `AND`
            ands = ' ( ' + ' AND '.join(ands) + ' ) '
            ors.append(ands)
        # join multiple qualifier sets with `OR`
        return ' OR '.join(ors)

    def __str__(self):
        """Return the cypher edge reference."""
        elements = [
            f":{self.label}" if self.label else "",
        ]
        if not self.length == (1, 1):
            elements.append("*{}..{}".format(
                self.length[0] if self.length[0] is not None else "",
                self.length[1] if self.length[1] is not None else "",
            ))
        elements.append(f"{self.prop_string}")
        return "-[`{0}`{1}]-".format(
            self.name,
            "".join(elements),
        ) + (">" if self.directed else "")


def build_match_clause(
        *patterns,
        filters=None,
        hints=None,
        qualifier_filters="",
        **kwargs,
):
    """Build MATCH clause (and subclauses) from components."""
    query = ""
    query += "MATCH " + ", ".join(patterns)
    reg_filters_cypher = ""
    qualifier_filters_cypher = ""
    combine_op = ""
    has_filters = False
    if kwargs.get("use_hints", False) and hints:
        query += " " + " ".join(hints)
    if filters or qualifier_filters:
        if len(filters):
            has_filters = True
            if len(filters) > 1:
                filters = [f"({f})" for f in filters]
                reg_filters_cypher += " ( " + " AND ".join(filters) + " ) "
            else:
                reg_filters_cypher += " ( " + filters[0] + " ) "

        if len(qualifier_filters):
            has_filters = True
            qualifier_filters_cypher += f"({qualifier_filters})"

        if len(filters) and len(qualifier_filters):
            combine_op = " AND "
    if has_filters:
        query += " WHERE " + reg_filters_cypher + combine_op + qualifier_filters_cypher

    return query


def match_edge(
    qedge_id,
    qedge,
    node_references: Dict[str, NodeReference],
    invert=True,
    **kwargs,
):
    """Get MATCH clause for edge."""
    eref = EdgeReference(qedge_id, qedge, invert=invert, **kwargs)
    if eref.cypher_invert:
        source_node = node_references[qedge["object"]]
        target_node = node_references[qedge["subject"]]
    else:
        source_node = node_references[qedge["subject"]]
        target_node = node_references[qedge["object"]]
    pattern = f"{source_node}{eref}{target_node}"
    edge_filters = [
        f"({c})"
        for c in source_node.filters + target_node.filters + eref.filters
    ]
    return build_match_clause(
        pattern,
        hints=source_node.hints + target_node.hints,
        filters=edge_filters,
        qualifier_filters=eref.qualifier_filters,
        **kwargs,
    )


def match_query(qgraph, subclass=True, **kwargs):
    """Generate a Cypher MATCH clause.

    Returns the query fragment as a string.
    """

    qgraph_nodes = qgraph["nodes"]
    qgraph_edges = qgraph["edges"]

    # Subclass is just a flag that can turn subclassing off completely.
    # We also check to make sure there is at least one qedge, because we decided that queries with only nodes shouldn't
    # do subclassing. This is primarily because there's currently no great way to represent those results.
    if subclass and qgraph_edges:
        # We don't want to include subclass nodes in queries that explicitly ask for subclass relationships.
        # Find all the qnodes that have subclass or superclass edges connected to them, so we can handle excluding them.
        qnode_ids_with_hierarchy_edges = set()
        for qedge_id, qedge in qgraph_edges.items():
            predicates = qedge.get("predicates", None)
            if predicates and ("biolink:subclass_of" in predicates or "biolink:superclass_of" in predicates):
                qnode_ids_with_hierarchy_edges.add(qedge['subject'])
                qnode_ids_with_hierarchy_edges.add(qedge['object'])
        superclasses = {
            # make superclass nodes for the pinned nodes (except ones with explicit subclass edges attached)
            qnode_id + "_superclass": {
                "ids": qnode.pop("ids"),
                "categories": qnode.pop("categories", None),
                "_superclass": True
            }
            for qnode_id, qnode in qgraph_nodes.items()
            if qnode.get("ids", None) is not None and qnode_id not in qnode_ids_with_hierarchy_edges
        }
        if 'subclass_depth' in kwargs:
            subclass_depth = kwargs['subclass_depth']
            if not isinstance(subclass_depth, int):
                raise TypeError(f"Unsupported subclass_depth type: {type(subclass_depth).__name__}.")
            elif subclass_depth < 0:
                raise ValueError(f"Parameter subclass_depth must be a positive integer.")
            else:
                subclass_depth = kwargs['subclass_depth']
        else:
            subclass_depth = 1
        # make variable length subclass edges to all the superclass nodes
        subclass_edges = {
            qnode_id[:-11] + "_subclass_edge": {
                "subject": qnode_id[:-11],
                "object": qnode_id,
                "predicates": ["biolink:subclass_of"],
                "_length": (0, subclass_depth),
                "_invert": False,
                "_subclass": True
            }
            for qnode_id in superclasses
        }
        # add the sub/super class nodes and edges to the qgraph
        qgraph_nodes.update(superclasses)
        qgraph_edges.update(subclass_edges)

    # sets of ids
    defined_nodes = set(qgraph_nodes)
    defined_edges = set(qgraph["edges"])
    referenced_nodes = set(
        [e["subject"] for e in qgraph_edges.values()]
        + [e["object"] for e in qgraph_edges.values()]
    )

    # generate internal node and edge variable names
    node_references = {
        qnode_id: NodeReference(qnode_id, qnode, **kwargs)
        for qnode_id, qnode in qgraph_nodes.items()
    }
    for node_id in referenced_nodes - defined_nodes:  # reference-only nodes
        node_references[node_id] = MissingReference(node_id)

    clauses = []

    # match orphaned nodes
    for node_id in defined_nodes - referenced_nodes:
        clauses.append(build_match_clause(
            str(node_references[node_id]),
            hints=node_references[node_id].hints,
            filters=node_references[node_id].filters,
            **kwargs,
        ))

    # match edges
    for qedge_id, qedge in qgraph_edges.items():
        clauses.append(match_edge(
            qedge_id,
            qedge,
            node_references,
            **kwargs,
        ))

    return Query(
        " ".join(clauses),
        qids=defined_nodes | defined_edges,
        references=defined_nodes | referenced_nodes | defined_edges,
        qgraph=qgraph,
    )
